#ifndef _BUILDERTRANSLATIONS_HH
#define _BUILDERTRANSLATIONS_HH

#include <string>

static struct std::string g_builderTranslations[] = {
_("Pinot"),
_("Query:"),
_("Stored queries"),
_("Query parameters"),
_("Name:"),
_("New filter or range:"),
_("Query Text"),
_("Index results with label"),
_("Number of results:"),
_("Sort order:"),
_("Stemming:"),
_("Only index new results"),
_("Properties"),
_("Title:"),
_("Language:"),
_("MIME type:"),
_("Size:"),
_("Terms:"),
_("bytes"),
_("External index"),
_("Name:"),
_("..."),
_("Type:"),
_("Location:"),
_("Port:"),
_("Address:"),
_("Import URL"),
_("Import"),
_("Location:"),
_("Title:"),
_("Apply label:"),
_("Status"),
_("Helper application"),
_("Use this command for other documents of the same type."),
_("Pinot Preferences"),
_("These directories will be indexed and optionally monitored for changes:"),
_("Add"),
_("File patterns:"),
_("Add"),
_("Reset"),
_("Indexing"),
_("Labels are used to classify indexed documents:"),
_("Add"),
_("Remove"),
_("Labels"),
_("Direct connection to the Internet"),
_("Manual proxy configuration:"),
_("Address:"),
_("Type:"),
_("Port:"),
_("Network"),
_("Google API key:"),
_("Queries:"),
_("Enable search terms suggestion"),
_("Ignore robots.txt and Robots META tag"),
_("HTTP crawling:"),
_("New results:"),
_("Miscellaneous"),
_("There is no helper application defined for this document's type.\n"
"Please enter the command to open this document."),
""
};

#endif // _BUILDERTRANSLATIONS_HH
